/** @file wiicom_libvcp.c
 *  @brief Wiicom virtual com port library
 *
 *  Generic library for managing a virtual com port
 *
 *  @author Marco Malano
 *  @author WIICOM s.r.l
 *  @version   1.0.0
 *  @date      2012
 *  @copyright 2012 - Wiicom - p.iva IT10710600015 - All rights reserved.
 *  @attention This library depends on @b wiicom_libutil \n
 *  Every function that expects a result, in case of an error, set the Wiicom errno and print to stderr the error occurred.
 *  @note None.
 *  @todo
 *  @deprecated
 *  @bug No bugs found.
 */
#include "wiicom_libvcp.h"

#define VCP_NUM_MAX_TIME_WRITE 5
#define MAX_NUM_VCP 25

#if (defined (__WIN32__))
//====================================================================================================
//                                          WINDOWS SECTION
//====================================================================================================

static struct
{
	DCB attributes;
	HANDLE vcp_handle;
} glb_old_vcp_attributes[MAX_NUM_VCP];  // struttura che contiene i vecchi attributi della vcp, prima della connessione

static int glb_num_vcp_used=0; //numero di vcp aperte

// funzione locale che ripulisce la struttura che contiene i vecchi attributi della vcp
static void loc_init_old_vcp_attributes()
{
	uint8_t idx=0;
	for(idx=0;idx<MAX_NUM_VCP;idx++)
	{
		memset((void *) &(glb_old_vcp_attributes[idx].attributes),'\0',sizeof(DCB));
		glb_old_vcp_attributes[idx].vcp_handle=0;
	}
	glb_num_vcp_used=0;
}

// funzione locale che registra i vecchi attributidella vcp
static uint8_t loc_set_old_vcp_attributes(HANDLE vcp_handle,DCB *attributes)
{
	uint8_t b_error=FALSE;
	uint8_t idx=0;

	for(idx=0;idx<MAX_NUM_VCP && idx<glb_num_vcp_used && b_error==FALSE;idx++)
	{
		if(glb_old_vcp_attributes[idx].vcp_handle==vcp_handle)  b_error=TRUE;
	}

	if(b_error==FALSE)
	{
		uint8_t b_trovato=FALSE;

		for(idx=0;idx<MAX_NUM_VCP && b_trovato==FALSE;idx++)
			if(glb_old_vcp_attributes[idx].vcp_handle==0)   b_trovato=TRUE;

		if(b_trovato)
		{
			memcpy((void *) &(glb_old_vcp_attributes[idx].attributes),(void *)attributes,sizeof(DCB));
			glb_old_vcp_attributes[idx].vcp_handle=vcp_handle;
			glb_num_vcp_used++;
		}
		else
			b_error=TRUE;
	}

	return(!b_error);
}

// funzione locale che preleva i vecchi attributi della vcp
static uint8_t loc_get_old_vcp_attributes(HANDLE vcp_handle,DCB *attributes)
{
	uint8_t b_error=FALSE;
	uint8_t idx=0;

	if(vcp_handle==0)
		b_error=TRUE;
	else
	{
		uint8_t b_trovato=FALSE;
		for(idx=0;idx<MAX_NUM_VCP && idx<glb_num_vcp_used && b_trovato==FALSE;idx++)
		{
			if(glb_old_vcp_attributes[idx].vcp_handle==vcp_handle)
			{
				memcpy((void *)attributes,(void *) &(glb_old_vcp_attributes[idx].attributes),sizeof(DCB));
				b_trovato=TRUE;
			}
		}
		if(b_trovato==FALSE)   b_error=TRUE;
	}

	return(!b_error);
}

// funzione locale che preleva i vecchi attributi della vcp
static uint8_t loc_reset_old_vcp_attributes(HANDLE vcp_handle)
{
	uint8_t b_error=FALSE;
	uint8_t idx=0;

	if(vcp_handle==0)
		b_error=TRUE;
	else
	{
		uint8_t b_trovato=FALSE;
		for(idx=0;idx<MAX_NUM_VCP && idx<glb_num_vcp_used && b_trovato==FALSE;idx++)
		{
			if(glb_old_vcp_attributes[idx].vcp_handle==vcp_handle)
			{
				memset((void *) &(glb_old_vcp_attributes[idx].attributes),'\0',sizeof(DCB));
				glb_old_vcp_attributes[idx].vcp_handle=0;
				glb_num_vcp_used--;
				b_trovato=TRUE;
			}
		}
		if(b_trovato==FALSE)   b_error=TRUE;
	}

	return(!b_error);
}

uint8_t wiicom_libvcp_open(t_wiicom_libvcp_handle *fd,char *device,uint32_t vcp_baudrate,uint8_t vcp_parity,uint8_t vcp_stop2bit,uint8_t vcp_data_bits_size)
{
	uint8_t b_error=FALSE;
	uint16_t parity=0;
	uint32_t baud=0;
	t_wiicom_libvcp_handle fd_loc=0;

	*fd=0;

	// controllo il baud rate
	switch (vcp_baudrate)
	{
		case WIICOM_VCP_BAUDRATE_110:
			baud=CBR_110;
        break;
		case WIICOM_VCP_BAUDRATE_300:
			baud=CBR_300;
        break;
		case WIICOM_VCP_BAUDRATE_600:
			baud=CBR_600;
        break;
		case WIICOM_VCP_BAUDRATE_1200:
			baud=CBR_1200;
        break;
		case WIICOM_VCP_BAUDRATE_2400:
			baud=CBR_2400;
        break;
		case WIICOM_VCP_BAUDRATE_4800:
			baud=CBR_4800;
        break;
		case WIICOM_VCP_BAUDRATE_9600:
			baud=CBR_9600;
        break;
		case WIICOM_VCP_BAUDRATE_19200:
			baud=CBR_19200;
        break;
		case WIICOM_VCP_BAUDRATE_38400:
			baud=CBR_38400;
        break;
		case WIICOM_VCP_BAUDRATE_57600:
			baud=CBR_57600;
        break;
		case WIICOM_VCP_BAUDRATE_115200:
			baud=CBR_115200;
        break;
		case WIICOM_VCP_BAUDRATE_230400:
			baud=230400;
		break;
		case WIICOM_VCP_BAUDRATE_256000:
			baud=CBR_256000;
        break;

		default:
			//wiicom_util_set_errno(WIICOM_EVCPBAUD);
			//b_error=TRUE;
			baud=vcp_baudrate;
        break;
	}

	if(!b_error)
	{
		switch(vcp_data_bits_size)
		{
			case WIICOM_VCP_DATA_BITS_SIZE_5:
			case WIICOM_VCP_DATA_BITS_SIZE_6:
			case WIICOM_VCP_DATA_BITS_SIZE_7:
			case WIICOM_VCP_DATA_BITS_SIZE_8:
				b_error=FALSE;
                break;

			default:
				wiicom_util_set_errno(WIICOM_EVCPDATASIZE);
				b_error=TRUE;
                break;
		}
	}

	if(!b_error)
	{
		switch(vcp_parity)
		{
			case WIICOM_VCP_NOPARITY:
				parity=NOPARITY;
                break;
			case WIICOM_VCP_EVENPARITY:
				parity=EVENPARITY;
                break;
			case WIICOM_VCP_ODDPARITY:
				parity=ODDPARITY;
                break;

			default:
				wiicom_util_set_errno(WIICOM_EVCPPARITY);
				b_error=TRUE;
            break;
		}
	}

	if(!b_error)
	{
		fd_loc = CreateFileA(device,GENERIC_READ|GENERIC_WRITE,
							   0,                          /* no share  */
							   NULL,                       /* no security */
							   OPEN_EXISTING,
							   0,                          /* no threads */
							   NULL);                      /* no templates */

		if(fd_loc==INVALID_HANDLE_VALUE)
		{
			wiicom_util_set_errno(GetLastError());
			b_error=TRUE;
		}
		else
		{
			DCB port_settings;
			COMMTIMEOUTS Cptimeouts;

			memset(&port_settings, 0, sizeof(port_settings));
			port_settings.DCBlength = sizeof (DCB);

			// Get the default port setting information.
			if(!GetCommState(fd_loc, &port_settings))
			{
				wiicom_util_set_errno(errno);
                CloseHandle(fd_loc);
				b_error=TRUE;
			}
			else
			{
				if(!loc_set_old_vcp_attributes(fd_loc,&port_settings))
				{
					wiicom_util_set_errno(WIICOM_EVCPATTR);
                    CloseHandle(fd_loc);
					b_error=TRUE;
				}
			}

			if(!b_error)
			{
                memset(&port_settings, 0, sizeof(port_settings));
                port_settings.DCBlength = sizeof(port_settings);
                port_settings.BaudRate=baud;
                port_settings.ByteSize=vcp_data_bits_size;
                port_settings.Parity=(BYTE)parity;
                if(parity==NOPARITY)	port_settings.fParity=FALSE;
                else	port_settings.fParity=TRUE;
                port_settings.fOutX=FALSE;
                port_settings.fInX=FALSE;
                port_settings.fBinary=TRUE;
                port_settings.fNull=FALSE;

                if(vcp_stop2bit)
                    port_settings.StopBits=ONESTOPBIT;
                else
                    port_settings.StopBits=TWOSTOPBITS;

                if(!SetCommState(fd_loc, &port_settings))
                {
                    wiicom_util_set_errno(GetLastError());
                    CloseHandle(fd_loc);
					loc_reset_old_vcp_attributes(fd_loc);
                    b_error=TRUE;
                }

                if(!b_error)
                {
                    Cptimeouts.ReadIntervalTimeout         = MAXDWORD;
                    Cptimeouts.ReadTotalTimeoutMultiplier  = 0;
                    Cptimeouts.ReadTotalTimeoutConstant    = 0;
                    Cptimeouts.WriteTotalTimeoutMultiplier = 0;
                    Cptimeouts.WriteTotalTimeoutConstant   = 0;

                    if(!SetCommTimeouts(fd_loc, &Cptimeouts))
                    {
                        wiicom_util_set_errno(GetLastError());
                        CloseHandle(fd_loc);
                        loc_reset_old_vcp_attributes(fd_loc);
                        b_error=TRUE;
                    }
                    else
                        *fd=fd_loc;
                }
            }
		}
	}

	return(!b_error);
}

void wiicom_libvcp_close(t_wiicom_libvcp_handle fd)
{
	CloseHandle(fd);
	loc_reset_old_vcp_attributes((HANDLE)fd);
}

long wiicom_libvcp_read(t_wiicom_libvcp_handle fd, uint8_t *buffer, uint32_t buffer_size)
{
	long n;

	if(buffer_size>4096)  buffer_size = 4096;

	/* added the void pointer cast, otherwise gcc will complain about */
	/* "warning: dereferencing type-punned pointer will break strict aliasing rules" */

	ReadFile(fd, buffer, buffer_size, (LPDWORD)((void *)&n), NULL);
	if(n<0)
	{
		wiicom_util_set_errno(GetLastError());
	}
	return(n);
}


long wiicom_libvcp_write(t_wiicom_libvcp_handle fd, uint8_t *buffer, uint32_t buffer_size)
{
	uint8_t num_tentativi=VCP_NUM_MAX_TIME_WRITE;
	uint32_t bytesSended=0;
	DWORD dwBytesWrite = 0;

	while(bytesSended<buffer_size && num_tentativi>0)
	{
		if(!WriteFile(fd, (buffer+bytesSended), buffer_size-bytesSended, &dwBytesWrite, NULL))
		{
			wiicom_util_set_errno(GetLastError());
			return(-1);
		}
		num_tentativi--;
		bytesSended+=dwBytesWrite;
	}


	return((long)bytesSended);
}

void wiicom_libvcp_deinit()
{
	uint8_t idx=0;
	for(idx=0;idx<MAX_NUM_VCP;idx++)
        if(glb_old_vcp_attributes[idx].vcp_handle!=0)	CloseHandle(glb_old_vcp_attributes[idx].vcp_handle);

	loc_init_old_vcp_attributes();
}

#else
//====================================================================================================
//                                         UNIX LIKE SECTION
//====================================================================================================

static struct
{
	struct termios attributes;
	int vcp_handle;
} glb_old_vcp_attributes[MAX_NUM_VCP];  // struttura che contiene i vecchi attributi della vcp, prima della connessione

static int glb_num_vcp_used=0; //numero di vcp aperte

// funzione locale che ripulisce la struttura che contiene i vecchi attributi della vcp
static void loc_init_old_vcp_attributes()
{
	uint8_t idx=0;
	for(idx=0;idx<MAX_NUM_VCP;idx++)
	{
		memset((void *) &(glb_old_vcp_attributes[idx].attributes),'\0',sizeof(struct termios));
		glb_old_vcp_attributes[idx].vcp_handle=0;
	}
	glb_num_vcp_used=0;
}

// funzione locale che registra i vecchi attributidella vcp
static uint8_t loc_set_old_vcp_attributes(int vcp_handle,struct termios *attributes)
{
	uint8_t b_error=FALSE;
	uint8_t idx=0;

	for(idx=0;idx<MAX_NUM_VCP && idx<glb_num_vcp_used && b_error==FALSE;idx++)
	{
		if(glb_old_vcp_attributes[idx].vcp_handle==vcp_handle)  b_error=TRUE;
	}

	if(b_error==FALSE)
	{
		uint8_t b_trovato=FALSE;

		for(idx=0;idx<MAX_NUM_VCP && b_trovato==FALSE;idx++)
			if(glb_old_vcp_attributes[idx].vcp_handle==0)   b_trovato=TRUE;

		if(b_trovato)
		{
			memcpy((void *) &(glb_old_vcp_attributes[idx].attributes),(void *)attributes,sizeof(struct termios));
			glb_old_vcp_attributes[idx].vcp_handle=vcp_handle;
			glb_num_vcp_used++;
		}
		else
			b_error=TRUE;
	}

	return(!b_error);
}

// funzione locale che preleva i vecchi attributi della vcp
static uint8_t loc_get_old_vcp_attributes(int vcp_handle,struct termios *attributes)
{
	uint8_t b_error=FALSE;
	uint8_t idx=0;

	if(vcp_handle==0)
		b_error=TRUE;
	else
	{
		uint8_t b_trovato=FALSE;
		for(idx=0;idx<MAX_NUM_VCP && idx<glb_num_vcp_used && b_trovato==FALSE;idx++)
		{
			if(glb_old_vcp_attributes[idx].vcp_handle==vcp_handle)
			{
				memcpy((void *)attributes,(void *) &(glb_old_vcp_attributes[idx].attributes),sizeof(struct termios));
				b_trovato=TRUE;
			}
		}
		if(b_trovato==FALSE)   b_error=TRUE;
	}

	return(!b_error);
}

// funzione locale che preleva i vecchi attributi della vcp
static uint8_t loc_reset_old_vcp_attributes(int vcp_handle)
{
	uint8_t b_error=FALSE;
	uint8_t idx=0;

	if(vcp_handle==0)
		b_error=TRUE;
	else
	{
		uint8_t b_trovato=FALSE;
		for(idx=0;idx<MAX_NUM_VCP && idx<glb_num_vcp_used && b_trovato==FALSE;idx++)
		{
			if(glb_old_vcp_attributes[idx].vcp_handle==vcp_handle)
			{
				memset((void *) &(glb_old_vcp_attributes[idx].attributes),'\0',sizeof(struct termios));
				glb_old_vcp_attributes[idx].vcp_handle=0;
				glb_num_vcp_used--;
				b_trovato=TRUE;
			}
		}
		if(b_trovato==FALSE)   b_error=TRUE;
	}

	return(!b_error);
}

uint8_t wiicom_libvcp_open(t_wiicom_libvcp_handle *fd,char *device,uint32_t vcp_baudrate,uint8_t vcp_parity,uint8_t vcp_stop2bit,uint8_t vcp_data_bits_size)
{
	uint8_t b_error=FALSE;
	int fd_loc=0;
	uint32_t baud=0;
	uint32_t cs=0;
	struct termios newtio;
	struct termios oldtio;

	*fd=0;

	// controllo il baud rate
	switch (vcp_baudrate)
	{
		case WIICOM_VCP_BAUDRATE_50:
			baud=B50;
            break;
		case WIICOM_VCP_BAUDRATE_75:
			baud=B75;
            break;
		case WIICOM_VCP_BAUDRATE_110:
			baud=B110;
            break;
		case WIICOM_VCP_BAUDRATE_134:
			baud=B134;
            break;
		case WIICOM_VCP_BAUDRATE_150:
			baud=B150;
            break;
		case WIICOM_VCP_BAUDRATE_200:
			baud=B200;
            break;
		case WIICOM_VCP_BAUDRATE_300:
			baud=B50;
            break;
		case WIICOM_VCP_BAUDRATE_600:
			baud=B600;
            break;
		case WIICOM_VCP_BAUDRATE_1200:
			baud=B1200;
            break;
		case WIICOM_VCP_BAUDRATE_1800:
			baud=B1800;
            break;
		case WIICOM_VCP_BAUDRATE_2400:
			baud=B2400;
            break;
		case WIICOM_VCP_BAUDRATE_4800:
			baud=B4800;
            break;
		case WIICOM_VCP_BAUDRATE_9600:
			baud=B9600;
            break;
		case WIICOM_VCP_BAUDRATE_19200:
			baud=B19200;
            break;
		case WIICOM_VCP_BAUDRATE_38400:
			baud=B38400;
            break;
		case WIICOM_VCP_BAUDRATE_57600:
			baud=B57600;
            break;
		case WIICOM_VCP_BAUDRATE_115200:
			baud=B115200;
            break;
		case WIICOM_VCP_BAUDRATE_230400:
			baud=B230400;
            break;

		default:
			//b_error=TRUE;
			//wiicom_util_set_errno(WIICOM_EVCPBAUD);
            baud=vcp_baudrate;
        break;
	}

	if(!b_error)
	{
		switch(vcp_data_bits_size)
		{
			case WIICOM_VCP_DATA_BITS_SIZE_5:
				cs=CS5;
                break;
			case WIICOM_VCP_DATA_BITS_SIZE_6:
				cs=CS6;
                break;
			case WIICOM_VCP_DATA_BITS_SIZE_7:
				cs=CS7;
                break;
			case WIICOM_VCP_DATA_BITS_SIZE_8:
				cs=CS8;
                break;

			default:
				wiicom_util_set_errno(WIICOM_EVCPDATASIZE);
				b_error=TRUE;
                break;
		}
	}

	if(!b_error)
	{
		switch(vcp_parity)
		{
			case WIICOM_VCP_NOPARITY:
			case WIICOM_VCP_EVENPARITY:
			case WIICOM_VCP_ODDPARITY:
				b_error=FALSE;
                break;

			default:
				wiicom_util_set_errno(WIICOM_EVCPPARITY);
				b_error=TRUE;
            break;
		}
	}

	if(!b_error)
	{
		fd_loc = open(device, O_RDWR | O_NOCTTY );
		if(fd_loc>0)
		{
			if(tcgetattr(fd_loc,&oldtio)==-1)
			{
				wiicom_util_set_errno(errno);
				close(fd_loc);
				b_error=TRUE;
			}
			else
			{
				if(!loc_set_old_vcp_attributes(fd_loc,&oldtio))
				{
					wiicom_util_set_errno(WIICOM_EVCPATTR);
					close(fd_loc);
					b_error=TRUE;
				}
			}

			if(!b_error)
			{
#if (defined (__APPLE__))
				cfmakeraw(&newtio);
				cfsetspeed(&newtio, baud);
				newtio.c_cflag = cs | CLOCAL | CREAD;
#else
				memset((void *)&newtio,0, sizeof(newtio));
				newtio.c_cflag = baud |  cs | CLOCAL | CREAD;
#endif
				newtio.c_cflag &= ~CRTSCTS;
				newtio.c_iflag &= ~(IXON | IXOFF | IXANY);
				switch(vcp_parity)
				{
					case WIICOM_VCP_NOPARITY:
						newtio.c_cflag &= ~PARENB;
                        break;
					case WIICOM_VCP_EVENPARITY:
						newtio.c_iflag |= PARENB;
                        break;
					case WIICOM_VCP_ODDPARITY:
						newtio.c_iflag |= PARODD;
                        break;
				}
				if(vcp_stop2bit)
					newtio.c_cflag |= CSTOPB;
				else
					newtio.c_cflag &= ~CSTOPB;

				newtio.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
				newtio.c_oflag=0;
				newtio.c_cc[VMIN] = 0;
				newtio.c_cc[VTIME] = 0;

				tcflush(fd_loc, TCIOFLUSH);
				//if(ioctl(fd_loc, TIOCSETA, &newtio)==-1)
				if(tcsetattr(fd_loc,TCSANOW,&newtio)==-1)
				{
					wiicom_util_set_errno(errno);
					loc_reset_old_vcp_attributes(fd_loc);
					close(fd_loc);
					b_error=TRUE;
				}
				else
					*fd=fd_loc;
			}
		}
		else
		{
			wiicom_util_set_errno(errno);
			b_error=TRUE;
		}
	}

	return(!b_error);
}

void wiicom_libvcp_close(t_wiicom_libvcp_handle fd)
{
	close(fd);
	loc_reset_old_vcp_attributes((int)fd);
}

long wiicom_libvcp_read(t_wiicom_libvcp_handle fd, uint8_t *buffer, uint32_t buffer_size)
{
	long n;

#ifndef __STRICT_ANSI__
	if(buffer_size>SSIZE_MAX)  buffer_size = (int)SSIZE_MAX;
#else
	if(buffer_size>4096)  buffer_size = 4096;
#endif

	n = read(fd, buffer, buffer_size);
	if(n<0)
	{
		if (EAGAIN != errno)	wiicom_util_set_errno(errno);
		else n=0;
	}
	return(n);
}

long wiicom_libvcp_write(t_wiicom_libvcp_handle fd, uint8_t *buffer, uint32_t buffer_size)
{
	uint8_t num_tentativi=VCP_NUM_MAX_TIME_WRITE;
	uint32_t bytesSended=0;
	long bytesWrite = 0;

	while(bytesSended<buffer_size && num_tentativi>0)
	{
//		uint8_t loop_wait=0;
	//	do
//		{
			bytesWrite=write(fd, (buffer+bytesSended), buffer_size-bytesSended);
//		}
//		while( bytesWrite<0 && EAGAIN == errno && loop_wait++<100);

		if(bytesWrite<0)
		{
				wiicom_util_set_errno(errno);
				return(-1);
		}
		num_tentativi--;
		bytesSended+=bytesWrite;
	}

	return((long)bytesSended);
}

void wiicom_libvcp_deinit()
{
	uint8_t idx=0;
	for(idx=0;idx<MAX_NUM_VCP;idx++)
		if(glb_old_vcp_attributes[idx].vcp_handle!=0)	close(glb_old_vcp_attributes[idx].vcp_handle);

	loc_init_old_vcp_attributes();
}

#endif

void wiicom_libvcp_init()
{
	loc_init_old_vcp_attributes();
}

long wiicom_libvcp_write_string(t_wiicom_libvcp_handle fd,char *string_to_write)
{
	long n;
    unsigned long len=strlen(string_to_write);
	n=wiicom_libvcp_write(fd,(uint8_t*)string_to_write, (uint32_t)len);

    return(n);
}


int wiicom_libvcp_getcomports(char ports[][15])
{
    int i=1;
    int j=0;
    int ret=0;

    t_wiicom_libvcp_handle fd_com_port=0;


#ifdef __WIN32__
    for(i=1; i<=255; i++)
    {
        char strPort[10] = "";
        sprintf (strPort, "\\\\.\\COM%d",i);

        ret = wiicom_libvcp_open(&fd_com_port,strPort,WIICOM_VCP_BAUDRATE_9600,WIICOM_VCP_NOPARITY,0,WIICOM_VCP_DATA_BITS_SIZE_8);

        if(ret)
        {
            char port[15] = "";
            sprintf (port, "COM%d",i);
            strcpy(ports[j],port);
            j++;
        }
        wiicom_libvcp_close(fd_com_port);
    }
#else
    FILE *fp;
    int status;
    char port[15];

    fp = popen("ls /sys/class/tty | grep -E 'ttyS|ttyUSB'", "r");
    if (fp == NULL) {
        return 0;
    }

    while (fgets(port, sizeof(port)-1, fp) != NULL) {
        if(port[strlen(port)-1]=='\n')
            port[strlen(port)-1]='\0';
        strcpy(ports[j], port);
        j++;
    }

    /* close */
    pclose(fp);
#endif

    return j;
}
